/**
 * Project: PS5
 * Authors: @Logan Tucker and @Tucker Simpson
 * Viterbi algorithm for cs10, winter 2022
 *
 */

import java.io.*;
import java.util.*;


public class Viterbi {


    /**
     * use viterbi to predict POS for every word in a string.
     * @param unseen log probability value for unseen use of a word
     * @param observations string of words
     * @param emissionMap map containing log probabilities of words being at different POSs
     * @param transmissionMap map containing log probabilities of different transitions between POSs
     * @return A list fo POS corresponding to each word in observations string
     */
    public static ArrayList<String> viterbiLine(int unseen, String observations, Map<String, HashMap<String, Double>> emissionMap,
                                                Map<String, HashMap<String, Double>> transmissionMap) {

        // key = nextState; value = currState
        Map<String, String> currStates = new HashMap<>();

        // currStates = { start }
        currStates.put("#", null);

        // key = currState; value = score
        Map<String, Double> currScores = new HashMap<>();

        // currScores = map { start=0 }
        currScores.put("#", 0.0);

        // key = curr, value = prev
        List<Map<String, String>> backPointers = new ArrayList<>();

        Map<String, String> nextStates = new HashMap<>();
        Map<String, Double> nextScores = new HashMap<>();

        // turn sentence into a list of words
        String[] obsList = observations.split(" ");
        List<String> obs = List.of(obsList);

        // for i from 0 to # observations - 1
        for (int i = 0; i < obs.size(); i++) {

            nextStates = new HashMap<>();
            nextScores = new HashMap<>();

            // for each currState in currStates
            for (String currState : currStates.keySet()) {

                // for each transition currState -> nextState
                for (String nextState : transmissionMap.get(currState).keySet()) {

                    // add nextState to nextStates
                    nextStates.put(nextState, currState);

                    double nextScore;

                    // set next-score for word as POS we have seen before
                    if (emissionMap.get(nextState).containsKey(obs.get(i))) {
                        nextScore = currScores.get(currState) + transmissionMap.get(currState).get(nextState) +
                                emissionMap.get(nextState).get(obs.get(i));
                    }

                    // set next-score for word as POS we have not seen before
                    else {
                        nextScore = currScores.get(currState) + transmissionMap.get(currState).get(nextState) + unseen;
                    }

                    // Add next score if there is not a score for that next state, or add if it is the best score for
                    // that state
                    if (!nextScores.containsKey(nextState) || nextScore > nextScores.get(nextState)) {
                        nextScores.put(nextState, nextScore);

                        if(backPointers.isEmpty()){
                            Map<String , String> iMap = new HashMap<String, String>();
                            iMap.put(nextState, currState);
                            backPointers.add(iMap);
                        }
                        if (backPointers.size() <= i) {
                            Map<String , String> iMap = new HashMap<String, String>();
                            iMap.put(nextState, currState);
                            backPointers.add(iMap);
                        } else {
                            backPointers.get(i).put(nextState, currState);
                        }
                    }
                }
            }

            currStates = nextStates;
            currScores = nextScores;
        }

        double currMax = Double.NEGATIVE_INFINITY;
        String bestState = "";

        // traverse through "graph" by highest scoring path to find likely last POS
        for(String state : nextStates.keySet()){
            if(nextScores.get(state) >currMax){
                currMax = currScores.get(state);
                bestState = state;
            }
        }

        // traverse back-pointers to from the likely last POS
        ArrayList<String> stateList = new ArrayList<>();
        for(int j = backPointers.size()-1; j>=0; j--){
            stateList.add(bestState);
            bestState = backPointers.get(j).get(bestState);
        }

        // flip the traversal list to return POS in order of word input
        ArrayList<String> rList = new ArrayList<>();
        for(int i = stateList.size()-1; i>=0; i--){
            rList.add(stateList.get(i));
        }

        return rList;
    }


    /**
     * Helper of userInput();
     * Takes a sentence and corresponding list of POS and formats them. (words/POS)
     * @param posList list of POS
     * @param obs sentence
     * @return string of words and POS
     * @throws Exception
     */
    public static String formatResults(ArrayList<String> posList, String obs) throws Exception {

        String[] obsList = obs.split(" ");

        String res = "";

        int i = 0;
        for (String ob : obsList) {
            res += ob + "/" + posList.get(i) + " ";
            i+=1;
        }
        res += "\n";
        return res;
    }


    /**
     * takes user input and uses viterbi to predict its POS
     * @param unseen log probability value for unseen use of a word
     * @param training
     * @throws Exception
     */
    public static void userInput(Integer unseen, Train training) throws Exception {

        Scanner myScanner = new Scanner(System.in);
        System.out.print("Enter Sentence: ");

        String input = myScanner.nextLine();

        System.out.println(formatResults(viterbiLine(unseen, input, training.EmissionMap, training.TransmissionMap), input));
    }


    /**
     * tests the accuracy of an outfile holding predicted POS versus an answer key file
     * @param myOutputFile file containing tags from using viterbi to detect POS
     * @param correctOutputFile file containing correct tags
     * @return String detailing how many correct and how many incorrect
     */
    public static String checkTest(String myOutputFile, String correctOutputFile){

        try {
            BufferedReader myOut = new BufferedReader(new FileReader(myOutputFile));
            BufferedReader corrOut = new BufferedReader(new FileReader(correctOutputFile));

            int correctCount = 0;
            int totalCount = 0;

            String myLine = myOut.readLine();
            String corrLine = corrOut.readLine();

            while(myLine!=null){
                String[] myPOS = myLine.split(" ");
                String[] corrPOS = corrLine.split(" ");

                for(int i = 0; i< myPOS.length; i++){
                    if(Objects.equals(myPOS[i], corrPOS[i])) correctCount +=1;
                    totalCount +=1;
                }
                myLine = myOut.readLine();
                corrLine = corrOut.readLine();
            }

            double correctPercent = 100.0 * correctCount/totalCount;

            return "Viterbi correctly tagged " + correctCount +"/" + totalCount + " parts of speech ("+correctPercent+
                    "% of words).";


        }catch (IOException e){
            System.out.println("Invalid Files");
        }

        return "Try again with different files";
    }


    /**
     * Writes POS predictions from an input file to an output file
     * @param unSeen log probability value for unseen use of a word
     * @param training an instance of train
     * @param testInput name of file containing test sentences
     * @param outputLocation location where output file will be written
     */
    public static void writeOutput(int unSeen, Train training, String testInput, String outputLocation) {
        try {

            BufferedReader input = new BufferedReader(new FileReader(testInput));

            BufferedWriter output = new BufferedWriter(new FileWriter(outputLocation));

            String line = input.readLine();

            while (line != null) {
                ArrayList<String> poss = viterbiLine(unSeen, line, training.EmissionMap, training.TransmissionMap);
                String rString = "";
                for (int i = 0; i < poss.size() - 1; i++) {
                    String currPos = poss.get(i);
                    rString += currPos + " ";
                }
                rString += poss.get(poss.size() - 1) + "\n";

                output.write(rString);

                line = input.readLine();

            }
            output.close();

        }catch(IOException e){
            System.out.println("Invalid Files");
        }
    }

    public static double logProbGetter(double num, double den){
        return Math.log(num/den);

    }


    public static void main(String[] args) throws Exception {
        String trainSentences = "/Users/tuckersimpson/IdeaProjects/cs10/ps5_inputs/brown-train-sentences.txt";
        String trainTags = "/Users/tuckersimpson/IdeaProjects/cs10/ps5_inputs/brown-train-tags.txt";
        String testSentences = "/Users/tuckersimpson/IdeaProjects/cs10/ps5_inputs/brown-test-sentences.txt";
        String testTags = "/Users/tuckersimpson/IdeaProjects/cs10/ps5_inputs/brown-test-tags.txt";
        String outputLocation = "/Users/tuckersimpson/IdeaProjects/cs10/ps5_outputs/brown-test-results.txt";

        int unseenPunishment = -100;

        Train training = new Train(trainSentences, trainTags);

        writeOutput(unseenPunishment, training, testSentences, outputLocation);

        System.out.println(checkTest(outputLocation, testTags));

        userInput(unseenPunishment, training);


        // HARD CODE TESTING
//        HashMap<String, HashMap<String, Double>> hardCodeEmissions = new HashMap<>();
//
//        HashMap<String, Double> internalMap = new HashMap<>();
//        internalMap.put("chase", logProbGetter(10.0, 10.0));
//        hardCodeEmissions.put("NP", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("and", logProbGetter(10.0, 10.0));
//        hardCodeEmissions.put("CNJ", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("get", logProbGetter(1.0, 10.0));
//        internalMap.put("chase", logProbGetter(3.0, 10.0));
//        internalMap.put("watch", logProbGetter(6.0, 10.0));
//        hardCodeEmissions.put("V", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("cat", logProbGetter(4.0, 10.0));
//        internalMap.put("dog", logProbGetter(4.0, 10.0));
//        internalMap.put("watch", logProbGetter(2.0, 10.0));
//        hardCodeEmissions.put("N", internalMap);
//
//        HashMap<String, HashMap<String, Double>> hardCodeTransmissions = new HashMap<>();
//        internalMap = new HashMap<>();
//        internalMap.put("NP", logProbGetter(3.0, 10.0));
//        internalMap.put("N", logProbGetter(7.0, 10.0));
//        hardCodeTransmissions.put("#", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("CNJ", logProbGetter(2.0, 10.0));
//        internalMap.put("V", logProbGetter(8.0, 10.0));
//        hardCodeTransmissions.put("NP", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("CNJ", logProbGetter(2.0, 10.0));
//        internalMap.put("V", logProbGetter(8.0, 10.0));
//        hardCodeTransmissions.put("N", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("N", logProbGetter(4.0, 10.0));
//        internalMap.put("V", logProbGetter(4.0, 10.0));
//        internalMap.put("NP", logProbGetter(2.0, 10.0));
//        hardCodeTransmissions.put("CNJ", internalMap);
//
//        internalMap = new HashMap<>();
//        internalMap.put("N", logProbGetter(4.0, 10.0));
//        internalMap.put("CNJ", logProbGetter(2.0, 10.0));
//        internalMap.put("NP", logProbGetter(4.0, 10.0));
//        hardCodeTransmissions.put("V", internalMap);
//
//        System.out.println(hardCodeEmissions);
//        System.out.println(hardCodeTransmissions);
//
//        String testObservation = "Logan ran and jumped";
//
//        System.out.println(viterbiLine(-100, testObservation, hardCodeEmissions, hardCodeTransmissions));

    }

}



